'
' Copies file/folder timestamps (created, modified, last accessed)
'
' last access timestamp copy does not work on windows xp (also via smb) as it gets updated
'
' Stefan Wurzinger, 2014-07-30
'

#Define UNICODE

#Include "windows.bi"
#Include "win/shellapi.bi"		'CommandLineToArgvW


'named CommandLineToArgv as of FB 0.90.1
#Ifndef CommandLineToArgvW
#Inclib "shell32"
'LPWSTR* CommandLineToArgvW(
'  _In_   LPCWSTR lpCmdLine,
'  _Out_  int *pNumArgs
');
Declare Function CommandLineToArgvW Alias "CommandLineToArgvW" (lpCmdLine As LPCWSTR, pNumArgs As Integer Ptr) As LPWSTR Ptr
#EndIf


#Define MAXPATH 2048


Dim As LPTSTR CmdLine = GetCommandLine()
Dim As Integer ArgCnt
Dim Shared As LPWSTR Ptr ArgList
ArgList = CommandLineToArgv(CmdLine, @ArgCnt) 




'translated from C, see http://stackoverflow.com/a/2315808
'added callback function argument
Function EnumerateDirectory(sDir As WString Ptr, callback As Function (path As WString Ptr, fdFile As WIN32_FIND_DATA Ptr) As Integer = 0) As Integer
	
	Dim fdFile As WIN32_FIND_DATA
	Dim hFind As HANDLE
	
	Dim sPath As WString*MAXPATH
	sPath = *sDir + WStr("\*")
	
	hFind = FindFirstFile(sPath, @fdFile)
	If hFind = INVALID_HANDLE_VALUE Then
		Print "Path not found: "; *sDir
		Return 0
	EndIf
	
	Do
		'Find first file will always return "." and ".." as the first two directories.
		If fdFile.cFileName <> WStr(".") AndAlso fdFile.cFileName <> WStr("..") Then
			
			'Build up our file path using the passed in 
      '[sDir] and the file/foldername we just found: 
			sPath = *sDir + WStr("\") + fdFile.cFileName
			
			'Is the entity a File or Folder?
			If fdFile.dwFileAttributes And FILE_ATTRIBUTE_DIRECTORY Then
				If callback <> 0 Then
					If callback(sPath, @fdFile) Then
						EnumerateDirectory(sPath, callback)		'//Recursion, I love it!
					EndIf
				Else
					Print WStr("Directory: "); sPath
					EnumerateDirectory(sPath, callback)		'//Recursion, I love it!
				EndIf
			Else
				If callback <> 0 Then
					callback(sPath, @fdFile)
				Else
					Print WStr("File: "); sPath
				End If
			EndIf
			
		EndIf
		
	Loop While FindNextFile(hFind, @fdFile)
	
  FindClose(hFind)			'Always, Always, clean things up!
  
  Return -1
  
End Function



' Copy create + access + write timestamps from directory -> directory or file -> file
Function CopyDates(src As WString Ptr, dst As WString Ptr) As Integer
	Dim As DWORD fileAttSrc = GetFileAttributes(*src)
	Dim As DWORD fileAttDst = GetFileAttributes(*dst)
	
	'e.g. if file does not exist or is not accessible
	If fileAttSrc = INVALID_FILE_ATTRIBUTES Or fileAttDst = INVALID_FILE_ATTRIBUTES Then Return 0
	
	'check for matching type (directory -> directory or file -> file)
	'If IIf(fileAttSrc And FILE_ATTRIBUTE_DIRECTORY, 1, 0) <> IIf(fileAttDst And FILE_ATTRIBUTE_DIRECTORY, 1, 0) Then Return 0
	
	Dim timeCreate As FILETIME
	Dim timeAccess As FILETIME
	Dim timeWrite As FILETIME
	
	Dim hDir As HANDLE
	
	'read dates
	hDir = CreateFile(src, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, IIf(fileAttSrc And FILE_ATTRIBUTE_DIRECTORY, FILE_ATTRIBUTE_DIRECTORY, 0) or FILE_FLAG_BACKUP_SEMANTICS, NULL)
	If hDir = INVALID_HANDLE_VALUE Then Return 0
	If GetFileTime(hDir, @timeCreate, @timeAccess, @timeWrite) = 0 Then Return 0
	CloseHandle(hDir)
	
	'Print
	'Print *cptr(LongInt ptr, @timeCreate); "  ";  *cptr(LongInt ptr, @timeAccess); "  "; *cptr(LongInt ptr, @timeWrite)
	
	'write dates
	'access errors when using GENERIC_WRITE, FILE_SHARE_WRITE; using FILE_SHARE_READ seems to work and fix the issue
	hDir = CreateFile(dst, GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_EXISTING, IIf(fileAttDst And FILE_ATTRIBUTE_DIRECTORY, FILE_ATTRIBUTE_DIRECTORY, 0) or FILE_FLAG_BACKUP_SEMANTICS, NULL)
	If hDir = INVALID_HANDLE_VALUE Then Return 0
	If SetFileTime(hDir, @timeCreate, @timeAccess, @timeWrite) = 0 Then Return 0
	CloseHandle(hDir)
	
	Return -1
	
End Function





Function CopyDateCallback(src As WString Ptr, fdFile As WIN32_FIND_DATA Ptr) As Integer
	Dim dst As WString*MAXPATH
	dst = *ArgList[2] + Mid(*src, Len(*ArgList[1])+1)
	Print *src; WStr(" -> "); dst; WStr(" : ");
	If CopyDates(src, dst) Then Print WStr("Success") Else Print WStr("Error") ': Return 0
	Return -1	'continue recursive enumeration
End Function





' first arg = app-name + 2 arguments
If ArgCnt = 3 Then
	
	Print *ArgList[1]; WStr(" -> "); *ArgList[2]; WStr(" : ");
	If CopyDates(*ArgList[1], *ArgList[2]) Then Print WStr("Success") Else Print WStr("Error")

' first arg = app-name + 3 arguments
ElseIf ArgCnt = 4 AndAlso *ArgList[3] = WStr("r") Then
	
	Print *ArgList[1]; WStr(" -> "); *ArgList[2]; WStr(" : ");
	If CopyDates(*ArgList[1], *ArgList[2]) Then Print WStr("Success") Else Print WStr("Error")
	EnumerateDirectory(*ArgList[1], @CopyDateCallback)

Else
	
	Print
	Print WStr(" Usage: copyDates <sourceDir> <targetDir> [r]")
	Print WStr("    or  copyDates <sourceFile> <targetFile> [r]")
	Print
	
EndIf


LocalFree(ArgList)



'EnumerateDirectory(WStr("C:\Users\Stefan\Desktop\FB Test\setdate"))
